/*
 * File:     WAVSampler.java
 * Project:  MPI Linguistic Application
 * Date:     15 August 2005
 *
 * Copyright (C) 2005  Max Planck Institute for Psycholinguistics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package mpi.eudico.client.util;

import java.io.IOException;
import java.io.RandomAccessFile;

import java.util.Arrays;


/**
 * A first shot at a class that provides samples given a certain media file. In
 * this version only .wav files are supported. In the future this class should
 * become more general and belongs in a different package than the current
 * one.
 *
 * @author Albert Russel, Alexander Klassmann
 * @version dec-2003
 */
public class WAVSampler {
    private String fileName;
    private RandomAccessFile soundFile;
    private byte[] buffer;
    private int[] intArrayFirst;
    private int[] intArraySecond;
    private int headerSize;
    private short nrOfChannels;
    private short bitrate;
    private short sampleSize;
    private int sampleFrequency;
    private float duration;
    private long nrOfSamples;
    private int possibleMinSample;
    private int possibleMaxSample;
    private int minSampleFirst;
    private int maxSampleFirst;
    private int minSampleSecond;
    private int maxSampleSecond;
    private short compressionCode;
    private WAVHeader wavHeader;

    /**
     * Construct the Samples object only for .wav files, otherwise an
     * IOException is thrown.
     *
     * @param fileName DOCUMENT ME!
     *
     * @throws IOException DOCUMENT ME!
     */
    public WAVSampler(String fileName) throws IOException {
        this.fileName = fileName;
        buffer = new byte[4096];

        if (fileName.toLowerCase().endsWith(".wav")) {
            soundFile = new RandomAccessFile(fileName, "r");
            wavHeader = new WAVHeader(soundFile);
            sampleFrequency = wavHeader.getFrequency();
            sampleSize = wavHeader.getSampleSize();
            nrOfSamples = wavHeader.getDataLength() / sampleSize;
            nrOfChannels = wavHeader.getNumberOfChannels();
            bitrate = (short) ((sampleSize * 8) / nrOfChannels);
            duration = ((float) 1000 * nrOfSamples) / sampleFrequency;
            possibleMinSample = (int) (-Math.pow(2, bitrate - 1));
            possibleMaxSample = (int) (-1 + Math.pow(2, bitrate - 1));
            headerSize = wavHeader.getHeaderSize();
            compressionCode = wavHeader.getCompressionCode();

            if (compressionCode == WAVHeader.WAVE_FORMAT_ALAW) {
                possibleMinSample *= 64;
                possibleMaxSample *= 64;
            }

            System.out.println("Information from header of wav-file:");
            System.out.println("NrOfChannels    : " + nrOfChannels);
            System.out.println("Sample frequency: " + sampleFrequency);
            System.out.println("Bitrate         : " + bitrate);
            System.out.println("nrOfSamples     : " + nrOfSamples);
            System.out.println(" -> Length      : " +
                TimeFormatter.toString(getTimeAtSample(nrOfSamples)));
            System.out.println("WAVE Format     : " +
                ((compressionCode < WAVHeader.formatDescriptions.length)
                ? WAVHeader.formatDescriptions[compressionCode]
                : ("" + compressionCode)));

            if (!wavHeader.getInfo().equals("")) {
                System.out.println("\nMeta info tail:" + wavHeader.getInfo());
            }

            soundFile.seek(headerSize);
        } else {
            throw new IOException("Unsupported file format");
        }
    }

    /**
     * Returns the sample frequency
     *
     * @return DOCUMENT ME!
     */
    public int getSampleFrequency() {
        return sampleFrequency;
    }

    /**
     * Returns the minimal possible value of the samples.
     *
     * @return DOCUMENT ME!
     */
    public int getPossibleMinSample() {
        return possibleMinSample;
    }

    /**
     * Returns the maximal possible value of the samples.
     *
     * @return DOCUMENT ME!
     */
    public int getPossibleMaxSample() {
        return possibleMaxSample;
    }

    /**
     * Returns the minimal value of the read samples.
     *
     * @return DOCUMENT ME!
     */
    public int getMinSampleFirst() {
        return minSampleFirst;
    }

    /**
     * Returns the maximal value of the read samples.
     *
     * @return DOCUMENT ME!
     */
    public int getMaxSampleFirst() {
        return maxSampleFirst;
    }

    /**
     * DOCUMENT ME!
     *
     * @return the minimal value of the read samples of the second channel
     */
    public int getMinSampleSecond() {
        return minSampleSecond;
    }

    /**
     * DOCUMENT ME!
     *
     * @return the mamimal value of the read samples of the second channel
     */
    public int getMaxSampleSecond() {
        return maxSampleSecond;
    }

    /**
     * returns time corresponding to a given sample
     *
     * @param sample
     *
     * @return long time in ms
     */
    public long getTimeAtSample(long sample) {
        return (long) (((sample * 1000f) / sampleFrequency) + .5);
    }

    /**
     * Seek the n-th sample
     *
     * @param n DOCUMENT ME!
     */
    public void seekSample(long n) {
        if ((n < 0) || (n > nrOfSamples)) {
            System.out.println("Cannot seek sample " + n + ".");
        } else {
            try {
                soundFile.seek(headerSize + (n * sampleSize));
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Get the current sample nr
     *
     * @return DOCUMENT ME!
     */
    public long getSamplePointer() {
        long pointer = 0;

        try {
            pointer = (soundFile.getFilePointer() - headerSize) / sampleSize;
        } catch (IOException e) {
            e.printStackTrace();
        }

        return pointer;
    }

    /**
     * Seek the sample that corresponds with the time in milli seconds
     *
     * @param time DOCUMENT ME!
     */
    public void seekTime(float time) {
        if (time < 0) {
            System.out.println("Cannot seek sample for time `" + time + "' !");
        } else {
            seekSample((long) ((time * sampleFrequency) / 1000));
        }
    }

    /**
     * Get the time in milli seconds that corresponds with the current sample
     *
     * @return DOCUMENT ME!
     */
    public float getTimePointer() {
        return ((float) 1000 * getSamplePointer()) / sampleFrequency;
    }

    /**
     *
     */
    public long getNrOfSamples() {
        return nrOfSamples;
    }

    /**
     * Returns the total duration of the samples in milli seconds
     *
     * @return DOCUMENT ME!
     */
    public float getDuration() {
        return duration;
    }

    /**
     * Reads an array of sample values
     *
     * @param samples the array that must be filled with samples
     *
     * @return DOCUMENT ME!
     *
     * @see #read(int[], int)
     */
    public int read(int[] samples) {
        return read(samples, samples.length);
    }

    /**
     * Read the specified number of samples into an (internal) int array. The
     * minimum and maximum values of the specified range of samples  can then
     * be retrieved by getMinSampleFirst and getMaxSampleFirst.<br>
     * If this method is called while the sound file contains more than  one
     * channel, the average sample value of the channels is calculated.
     *
     * @param nrOfSamplesToRead the number of samples to read
     *
     * @return the number of samples that actually have been read
     *
     * @see #read(int[], int)
     */
    public int read(int nrOfSamplesToRead) {
        if ((intArrayFirst == null) ||
                (intArrayFirst.length < nrOfSamplesToRead)) {
            intArrayFirst = new int[nrOfSamplesToRead];
        }

        return read(intArrayFirst, nrOfSamplesToRead);
    }

    /**
     * Read the specified number of samples into two (internal) int arrays. The
     * minimum and maximum values for each channel of the specified range  can
     * then be retrieved by getMinSampleFirst and getMaxSampleFirst for the
     * first channel and getMinSampleSecond and getMaxSampleSecond for the
     * second channel.<br>
     * Note: there is no support for audio files with more than two channels.
     *
     * @param nrOfSamplesToRead the number of files to read
     * @param nrOfChannelsToLoad the number of channels to read, either 1 or 2
     *
     * @return the number of samples that actually have been read
     *
     * @see #doRead()
     */
    public int read(int nrOfSamplesToRead, int nrOfChannelsToLoad) {
        if ((intArrayFirst == null) ||
                (intArrayFirst.length < nrOfSamplesToRead)) {
            intArrayFirst = new int[nrOfSamplesToRead];
        }

        if (nrOfChannelsToLoad == 2) {
            if ((intArraySecond == null) ||
                    (intArraySecond.length < nrOfSamplesToRead)) {
                intArraySecond = new int[nrOfSamplesToRead];
            }
        } else {
            nrOfChannelsToLoad = 1;
            intArraySecond = null;
        }

        return doRead(intArrayFirst, intArraySecond, nrOfSamplesToRead,
            nrOfChannelsToLoad);
    }

    /**
     * Reads an array of sample values.
     *
     * @param samples the array that must be filled with samples
     * @param nrOfSamplesToRead the number of samples that must be read into
     *        the array
     *
     * @return DOCUMENT ME!
     *
     * @see #doRead()
     */
    public int read(int[] samples, int nrOfSamplesToRead) {
        return doRead(samples, null, nrOfSamplesToRead, 1);
    }

    /**
     * Actually reads the samples from the soundfile. When the soundfile
     * contains only one channel, the array for the first channel is used.
     *
     * @param samples array to store the values of the first channel
     * @param samples2 array to store the values of the second channel
     * @param nrOfSamplesToRead the number of samples to read
     * @param nrOfChannelsToLoad the number of channels to load (seperately)
     *
     * @return the number of samples that actually have been read
     */
    private int doRead(int[] samples, int[] samples2, int nrOfSamplesToRead,
        int nrOfChannelsToLoad) {
        boolean stereo = false;

        try {
            if ((nrOfChannelsToLoad == 2) && (samples2 != null)) {
                stereo = true;
            }

            Arrays.fill(samples, (int) 0);

            if (samples2 != null) {
                Arrays.fill(samples2, (int) 0);
            }

            if (nrOfSamplesToRead > samples.length) {
                nrOfSamplesToRead = samples.length;
            }

            int samplesAvailable = (int) (getNrOfSamples() -
                getSamplePointer());

            if (nrOfSamplesToRead > samplesAvailable) {
                nrOfSamplesToRead = samplesAvailable;
            }

            if (buffer.length < (nrOfSamplesToRead * sampleSize)) {
                buffer = new byte[nrOfSamplesToRead * sampleSize];
            }

            if (sampleSize == 1) {
                soundFile.read(buffer, 0, nrOfSamplesToRead * sampleSize);

                for (int s = 0; s < nrOfSamplesToRead; s++) {
                    samples[s] = convert(buffer[s]);
                }
            } else if (sampleSize == 2) {
                soundFile.read(buffer, 0, nrOfSamplesToRead * sampleSize);

                int b = 0;
                int b1;
                int b2;

                for (int s = 0; s < nrOfSamplesToRead; s++) {
                    b1 = convert(buffer[b] & 0xFF);
                    b2 = convert(buffer[b + 1] & 0xFF);
                    samples[s] = (int) ((short) (b1 | (b2 << 8)));
                    b += 2;
                }
            } else if ((sampleSize == 4) && (nrOfChannels == 2)) {
                soundFile.read(buffer, 0, nrOfSamplesToRead * sampleSize);

                int b = 0;
                int b1;
                int b2;

                for (int s = 0; s < nrOfSamplesToRead; s++) {
                    // channel 1
                    b1 = convert(buffer[b] & 0xFF);
                    b2 = convert(buffer[b + 1] & 0xFF);
                    samples[s] = (int) ((short) (b1 | (b2 << 8)));

                    // channel 2
                    b1 = convert(buffer[b + 2] & 0xFF);
                    b2 = convert(buffer[b + 3] & 0xFF);

                    if (stereo) {
                        samples2[s] = (int) ((short) (b1 | (b2 << 8)));
                    } else {
                        //samples[s] += (int) ((short) (b1 | (b2 << 8)));//old
                        //samples[s] = (samples[s] + (int) ((short) (b1 | (b2 << 8)))) / 2;
                        samples[s] = Math.round((samples[s] +
                                (int) ((short) (b1 | (b2 << 8)))) / 2);
                    }

                    b += 4;
                }
            }
        } catch (IOException ioe) {
            System.out.println("Error while reading samples.");
            ioe.printStackTrace();
        }

        minSampleFirst = samples[0];
        maxSampleFirst = samples[0];

        for (int j = 1; j < nrOfSamplesToRead; j++) {
            if (samples[j] < minSampleFirst) {
                minSampleFirst = samples[j];
            } else if (samples[j] > maxSampleFirst) {
                maxSampleFirst = samples[j];
            }
        }

        if (stereo) {
            minSampleSecond = samples2[0];
            maxSampleSecond = samples2[0];

            for (int j = 1; j < nrOfSamplesToRead; j++) {
                if (samples2[j] < minSampleSecond) {
                    minSampleSecond = samples2[j];
                } else if (samples2[j] > maxSampleSecond) {
                    maxSampleSecond = samples2[j];
                }
            }
        }

        return nrOfSamplesToRead;
    }

    /**
     * Reads the requested number of samples in one run and stores the
     * extracted values  in one (mono or merged) or two (stereo) arrays of
     * ints. These arrays can then  be accessed to further process the samples.<br>
     * <b>Note:</b> if the loading of two channels is requested, a second array
     * of the proper size  will always be created, even if the .wav file does
     * not actually contain more than one channel.
     *
     * @param requestedNrOfSamplesToRead the number of samples to read
     * @param nrOfChannelsToLoad the number of channels to process from the
     *        file
     *
     * @return the number of <b>samples</b> that have been read, less than or
     *         equal to requestedNrOfSamplesToRead
     *
     * @see #getFirstChannelArray()
     * @see #getSecondChannelArray()
     */
    public int readInterval(int requestedNrOfSamplesToRead,
        int nrOfChannelsToLoad) {
        int actualRead = 0;
        boolean stereo = false;

        int actualNrOfSamplesToRead = requestedNrOfSamplesToRead;

        int samplesAvailable = (int) (getNrOfSamples() - getSamplePointer());

        if (requestedNrOfSamplesToRead > samplesAvailable) {
            actualNrOfSamplesToRead = samplesAvailable;
        }

        if ((intArrayFirst == null) ||
                (intArrayFirst.length < actualNrOfSamplesToRead)) {
            intArrayFirst = new int[actualNrOfSamplesToRead];
        }

        Arrays.fill(intArrayFirst, (int) 0);

        if (nrOfChannelsToLoad == 2) {
            stereo = true;

            if ((intArraySecond == null) ||
                    (intArraySecond.length < actualNrOfSamplesToRead)) {
                intArraySecond = new int[actualNrOfSamplesToRead];
            }

            Arrays.fill(intArraySecond, (int) 0);
        } else {
            nrOfChannelsToLoad = 1;
            intArraySecond = null;
        }

        if (buffer.length < (actualNrOfSamplesToRead * sampleSize)) {
            buffer = new byte[actualNrOfSamplesToRead * sampleSize];
        }

        // actual reading
        try {
            actualRead = soundFile.read(buffer, 0,
                    actualNrOfSamplesToRead * sampleSize);

            //System.out.println("Requested: " + requestedNrOfSamplesToRead);
            //System.out.println("Bytes read: " + actualRead);
        } catch (IOException ioe) {
            System.out.println("IO Error while reading samples.");

            //ioe.printStackTrace();
            return actualRead; // = 0
        }

        if (sampleSize == 1) {
            for (int s = 0; s < actualNrOfSamplesToRead; s++) {
                intArrayFirst[s] = convert(buffer[s] & 0xFF);
            }
        } else if (sampleSize == 2) {
            if (nrOfChannels == 1) {
                int b = 0;
                int b1;
                int b2;

                for (int s = 0; s < actualNrOfSamplesToRead; s++) {
                    b1 = buffer[b] & 0xFF;
                    b2 = buffer[b + 1] & 0xFF;
                    intArrayFirst[s] = convert((int) ((short) (b1 | (b2 << 8))));
                    b += 2;
                }
            } else {
                int b = 0;
                int b1;

                for (int s = 0; s < actualNrOfSamplesToRead; s++) {
                    // channel 1
                    b1 = buffer[b] & 0xFF;
                    intArrayFirst[s] = convert(b1);

                    // channel 2
                    b1 = buffer[b + 1] & 0xFF;

                    if (stereo) {
                        intArraySecond[s] = convert(b1);
                    } else {
                        intArrayFirst[s] = Math.round(((intArrayFirst[s] +
                                convert(b1)) / 2));
                    }

                    b += 2;
                }
            }
        } else if ((sampleSize == 4) && (nrOfChannels == 2)) {
            int b = 0;
            int b1;
            int b2;

            for (int s = 0; s < actualNrOfSamplesToRead; s++) {
                // channel 1
                b1 = buffer[b] & 0xFF;
                b2 = buffer[b + 1] & 0xFF;
                intArrayFirst[s] = convert((int) ((short) (b1 | (b2 << 8))));

                // channel 2
                b1 = buffer[b + 2] & 0xFF;
                b2 = buffer[b + 3] & 0xFF;

                if (stereo) {
                    intArraySecond[s] = convert((int) ((short) (b1 | (b2 << 8))));
                } else {
                    intArrayFirst[s] = Math.round((intArrayFirst[s] +
                            (convert((int) ((short) (b1 | (b2 << 8)))) / 2)));
                }

                b += 4;
            }
        }

        return actualRead / sampleSize;
    }

    /**
     * Gets the array that stores the data of the first channel of an interval
     * that has been read.
     *
     * @return an array of ints
     *
     * @see #readInterval(int, int)
     */
    public int[] getFirstChannelArray() {
        return intArrayFirst;
    }

    /**
     * Gets the array that stores the data of the second channel of an interval
     * that has been read.
     *
     * @return an array of ints
     *
     * @see #readInterval(int, int)
     */
    public int[] getSecondChannelArray() {
        return intArraySecond;
    }

    /**
     * Gets the <code>WavHeader</code>.
     *
     * @return the <code>WavHeader</code> object of this
     *         <code>WavSampler</code>
     */
    public WAVHeader getWavHeader() {
        return wavHeader;
    }

    /**
     * Close the RandomAccessFile.
     */
    public void close() {
        wavHeader = null;

        if (soundFile != null) {
            try {
                soundFile.close();
            } catch (IOException ioe) {
                ioe.printStackTrace();
            }
        }

        soundFile = null;
    }

    private int convert(int orig) {
        switch (compressionCode) {
        case WAVHeader.WAVE_FORMAT_ALAW: //alaw

            byte val = toUnsigned((byte) orig);
            int t;
            int seg;

            val ^= 0x55;
            t = (val & 0xf) << 4;
            seg = (toUnsigned(val) & 0x70) >> 4;

            switch (seg) {
            case 0:
                t += 8;

                break;

            case 1:
                t += 0x108;

                break;

            default:
                t += 0x108;
                t <<= (seg - 1);
            }

            return (((val & 0x80) != 0) ? t : (-t));

        default:
            return orig;
        }
    }

    private byte toUnsigned(byte signed) {
        return (byte) (signed + 128);
    }
}
